//////////////////  Closures eks 1  //////////////////////////////
/*
Funktionen greet er en funktion der returnere en ny funktion.
Funktionen greet er "parent function" til den anonyme funktion der returneres. Den anonyme funktions definition
er et function expression, da den returnerer et funktions objekt. 
*/
function greet(whatToSay) {
    return function(name) {
        console.log(whatToSay + ' ' + name);
    }
}

greet('Hi')('Henrik');      //output: Hi Henrik

/*
Når greet kaldes returneres en funktion, som igen kan kaldes - det gøres med de ekstra paranteser ( ) 
Vi kan også tildele funktionen der returneres fra greet til en variable og så kalde funktionen senere:
*/

var sayHi = greet('Hi');
sayHi('Henrik');            //output: Hi Henrik

/*
Det er fænomenet Closures der gør at whatToSay variablen stadigt lever og kan refereres fra den anonyme funktion
selv efter den ydre funktion (greet) er afsluttet. De variable der findes i Outer Environment bliver "Closed in" 
i Execution context for den anonyme funktion.
*/

//////////////////  Closures eks 2  //////////////////////////////


function buildFunctions(){
    var arr=[];
    for (var i=0; i<3; i++){
        arr.push(function() {console.log(i);})
    }
    return arr;
}

var fs = buildFunctions();
fs[0]();
fs[1]();
fs[2]();



function buildFunctions2(){
    var arr=[];
    for (var i=0; i<3; i++){
        let j = i;
        arr.push(function() {console.log(j);})
    }
    return arr;
}

var fs2 = buildFunctions2();
fs2[0]();
fs2[1]();
fs2[2]();


//////////////////  Closures eks 3 - Function Factories  //////////////////////////////

function makeGreeting(language) {
    return function(firstname, lastname){
        if (language==='en'){console.log('Hello ' + firstname + ' ' + lastname)};
        if (language==='dk'){console.log('Hej ' + firstname + ' ' + lastname)};
        if (language==='es'){console.log('Hola ' + firstname + ' ' + lastname)}
    }
}

var greetEnglish = makeGreeting('en');
var greetDanish = makeGreeting('dk');
var greetSpanish = makeGreeting('es');

greetEnglish('John', 'Doe');
greetDanish('Henrik', 'Høltzer');
greetSpanish('Pablo', 'Fuentes');



/*
Bemærk hvert kald af makeGreeting( ) resultere i et nyt Execution Context, det betyder at hvert kald 
får sin egen plads i hukommelsen. Functions expressions returnere et funktions objekt og Closures gør at 
den ydre variable "language" bevares og kan benyttes når den indre anonyme funktion kaldes
*/

// nu med ES6 Arrow-function i stedet for traditionel funktions definition ///////// Opgave Løsning ////////////
function makeGreeting2(language) {
    return (firstname, lastname) => {
        if (language==='en'){console.log('Hello ' + firstname + ' ' + lastname)};
        if (language==='dk'){console.log('Hej ' + firstname + ' ' + lastname)};
        if (language==='es'){console.log('Hola ' + firstname + ' ' + lastname)}
    }
}


//////////////////  Closures eks 4 - Function Factories  ///////// Opgave Løsning /////////////////////

function makeAdder(x) {
    return function(y) {
      return x + y;
    };
  }
  
  var add5 = makeAdder(5);
  var add10 = makeAdder(10);
  var addHello = makeAdder("Hello ");

  console.log(add5(2));  // 7
  console.log(add10(2)); // 12
  console.log(addHello("Henrik"));




//////////////////  Closures eks 5 - Callbacks  //////////////////////////////

function sayHiLater(){
    var greeting = 'Hi';
    setTimeout(function(){console.log(greeting);}, 3000)
}

sayHiLater();


/*
Den anonyme funktion der gives til setTimeout er en "First class function" (betyder at funktionen kan benyttes som 
argumenter og oprettes "on the fly"). Funktions definitionen i argumentet til setTimeout er et "function expressions" 

setTimeout er et asynkront kald der indgår i event-loopet - dvs setTimeout lukker og sendes til Browseren for at blive eksekveret.
Her ventes til tiden er gået (3000 ms) og der droppes et event. JS Engine checker om der er funktioner der venter/lytter og
kalder vores anonyme funktion. Da variablen greeting ikke findes lokalt i funktionen og sayHiLater er afsluttet, søges i "scope chain" 
og greeting findes i funktionens Closures. Så pga Closure fænomenet har funktionen fortsat adgang til greeting.
*/




///////////////////////////// Callback function /////////////////////////////////

function tellMeWhenDone(callback){
    // .... some work
    callback();  // 'callback': Når funktionen er færdig (med "some work") kaldes den funktion der blev givet som argument (callback funktionen)
}

tellMeWhenDone(function(){console.log("Im' done!")});
tellMeWhenDone(function(){console.log("All done!")});





//////////////////////////// Promise ///////////////////////
// copy past koden til devtool console og kør, prøv også at skrive reject()

var promise = new Promise((resolve, reject)  => {
    resolve(); // reject();
});

promise.then(()=>console.log('Im finish')); // her er registreret et callback der udføres når promise skifter status/tilsand til resolved

promise.then(()=>console.log('Im finish')).then(()=>console.log('jeg blev også kaldt')); // her er registreret 2 callbacks der begge kaldes

//dette kan skrives mere læsbart:
promise
    .then(()=>console.log('Im finish'))
    .then(()=>console.log('jeg blev også kaldt'))
    .catch(() => console.log('uh oh!!'))



//////////////////// More Promise /////////////////////////

promise = new Promise((resolve, reject)  => {
  setTimeout(()=>{resolve();}, 3000);
});

promise
    .then(()=>console.log('Im finish'))
    .then(()=>console.log('jeg blev også kaldt'))
    .catch(() => console.log('uh oh!!'))


///// Test med ande navne /////////////////// 
//det er ikke navnene: resolve og reject, men parameterlisten der betyder noget
promise = new Promise((res, rej)  => {
  setTimeout(()=>{res();}, 3000);
});

promise
    .then(()=>console.log('Im finish'))
    .then(()=>console.log('jeg blev også kaldt'))
    .catch(() => console.log('uh oh!!'))

   
//////////////////// More Promise - fetch /////////////////////////

url = "https://jsonplaceholder.typicode.com/posts/";

fetch(url)
  .then(data => console.log(data));

//Bemærk vi får ikke umiddelbart vores data tilbage med fetch,
//kun et objekt der repræsentere det responce vi får fra serveren.
//men vi kan få vores data ved at kalde json() først:

url = "https://jsonplaceholder.typicode.com/posts/";
fetch(url)
  .then(responce => responce.json())
  .then(data => console.log(data));


//En anden uhensigtsmæssighed ved fetch - catch udføres ikke ved fx 404 som forventet
//som det gøres i andre lib ala axios

url = "https://jsonplaceholder.typicode.com/posts123456/";

fetch(url)
  .then(data => console.log(data))
  .catch (error => console.log('BAD', error));

// Catch kaldes kun hvis fx serveren ikke findes:
url = "https://jsonplaceholder.typicode123456.com/posts/";

fetch(url)
  .then(data => console.log(data))
  .catch (error => console.log('BAD', error));

//Derfor anbefales axios i stedet





url = "https://api.icndb.com/jokes/random";
fetch(url)
  .then(responce => responce.json())
  .then(data => console.log(data))
  .catch (error => console.log('BAD', error));

url = "https://ergast.com/api/f1/2018/drivers.json";
fetch(url)
  .then(responce => responce.json())
  .then(data => console.log(data));
 
  

/////////////////////// promise - Med traditionelt AJAX ///////
promise = new Promise((resolve, reject) => {
  const request = new XMLHttpRequest();

  request.open('GET', 'https://api.icndb.com/jokes/random');
  request.onload = () => {
    if (request.status === 200) {
      resolve(request.response); // we got data here, so resolve the Promise
    } else {
      reject(Error(request.statusText)); // status is not 200 OK, so reject
    }
  };

  request.onerror = () => {
    reject(Error('Error fetching data.')); // error occurred, reject the  Promise
  };

  request.send(); // send the request
});

console.log('Asynchronous request made.');

promise.then((data) => {
  console.log('Got data! Promise fulfilled.');
  document.body.textContent = JSON.parse(data).value.joke;
}, (error) => {
  console.log('Promise rejected.');
  console.log(error.message);
});


//////////////////////////// Axios ////////////////////////

//import axios from 'axios';
//const axios = require('axios')
// <script src="https://unpkg.com/axios/dist/axios.min.js"></script>

/*
const getBreeds = () => {
  try {
    return axios.get('https://dog.ceo/api/breeds/list/all')
  } catch (error) {
    console.error(error)
  }
}

const countBreeds = async () => {
  const breeds = getBreeds()
    .then(response => {
      if (response.data.message) {
        console.log(
          `Got ${Object.entries(response.data.message).length} breeds`
        )
      }
    })
    .catch(error => {
      console.log(error)
    })
}

countBreeds()

*/

const getBreeds = async () => {
    try {
      return await axios.get('https://dog.ceo/api/breeds/list/all')
    } catch (error) {
      console.error(error)
    }
  }
  
  const countBreeds = async () => {
    const breeds = await getBreeds()
  
    if (breeds.data.message) {
      console.log(`Got ${Object.entries(breeds.data.message).length} breeds`)
    }
  }
  
  countBreeds()




  /////////////////////// axios - Med Promise //////////////////
  
  function getFormular1DriverInfo() {
    axios.get('https://ergast.com/api/f1/2018/drivers.json')
      .then(res => {
        console.log(res.data.MRData.DriverTable.Drivers);
      })
      .catch(err => {
        console.log(err);
      });
  }
  
  getFormular1DriverInfo();
  
  
  /////////////////////// axios - Med async/await //////////////////

  const getFomular1Drivers = async () => {
    try {
      return await axios.get('https://ergast.com/api/f1/2018/drivers.json')
    } catch (error) {
      console.error(error)
    }
  }
  
  const DriversInfo = async () => {
    let drivers = await getFomular1Drivers()
  
    console.log(drivers.data.MRData.DriverTable.Drivers);
    
  }
  
  DriversInfo()